import numpy as np
from random import sample
from masced_bandits.bandits.Bandit import Bandit
import matplotlib.pyplot as plt



#formula = "ao"
FORMULA_FUNC = None

REWARD = 0
ACTION = 1




class DUCB(Bandit):
    def __init__(self, **kwargs):
        gamma = float(kwargs.get("gamma",0.99))
        super().__init__("DUCB-" + str(gamma))

        self.discount = gamma
        self.bandit_round = -1
        self.game_list = []
        
    def get_next_arm(self, reward):
        self.game_list.append([reward, self.last_action])

        self.bandit_round = self.bandit_round + 1

        if((self.bandit_round) < (len(self.arms))):
            next_arm = self.arms[self.bandit_round]


        else:
            #scores = [str(self.reward_average(arm)) + " c: " + str(self.tuned(arm, sum([self.times_played(arm) for arm in self.arms]))) for arm in self.arms]
            #print("Scores were " + str(scores))
            next_arm = max(self.arms, key=lambda arm: \
                       self.reward_average(arm) + self.tuned(arm, sum([self.times_played(arm) for arm in self.arms])))


        self.last_action = next_arm
        return next_arm


    def reward_average(self, arm):
        r_sum = 0.0
        total_games = len(self.game_list)

        for i, game in enumerate(self.game_list):  
            if(game[ACTION] == arm): #the games in which arm was chosen
                #print("value added to sum: " + str(np.power(self.discount, total_games - i) * game[REWARD]))
                r_sum+=(np.power(self.discount, total_games - i) * game[REWARD])
     
        times_arm_played = self.times_played(arm) 
        if(r_sum == 0 or times_arm_played == 0): return 0
            
        return r_sum/times_arm_played

    
    def sqreward_average(self, arm):
        r_sum = 0.0
        total_games = len(self.game_list)
        for i, game in enumerate(self.game_list):  
            if(game[ACTION] == arm): #the games in which arm was chosen
                r_sum+=np.square(np.power(self.discount, total_games - i) * game[REWARD])

        times_arm_played = self.times_played(arm) 
        if(r_sum == 0 or times_arm_played == 0): return 0
            
        return r_sum/times_arm_played

    def times_played(self, arm):
        arm_played = 0.0
        total_games = len(self.game_list)
        for i, game in enumerate(self.game_list):
            if(game[ACTION] == arm): arm_played += np.power(self.discount, total_games - i) #Each time the g am


        return arm_played


    def tuned(self, arm, n):
        n_k = self.times_played(arm)

        average_of_squares = self.sqreward_average(arm)
        square_of_average = np.square(self.reward_average(arm))
        estimated_variance = average_of_squares - square_of_average
        param = np.log(n) / n_k
        
        V_k = estimated_variance + np.sqrt(2 * param)
        
        #print("values inside tuned " + str((n_k,average_of_squares,square_of_average,estimated_variance,param,V_k)))
        confidence_value = np.sqrt(param * V_k)
        if(np.isnan(confidence_value)):
            print("\n\n\n\ncaught a nan\n\n\n\n\n")
            return 0
        else:
            return confidence_value
    # def visualize(self):
    #     if((self.bandit_round) < (len(self.arms))): return
    #     arm_names = []
    #     arm_rewards= []
    #     arm_conf = []
    #     [(arm_names.append(str(arm)), arm_rewards.append(self.reward_average(arm)), arm_conf.append(self.tuned(arm, sum([self.times_played(arm) for arm in self.arms])))) for arm in self.arms]
    #     reward_bar = plt.bar(arm_names, arm_rewards)
    #     confidence_bar = plt.bar(arm_names, arm_conf, bottom=arm_rewards)
    #     #plt.bar_label(reward_bar, padding=3)
    #     #plt.bar_label(confidence_bar, padding=3)
    #     plt.yticks(np.arange(1.0,3.00, step=0.05))
    #     plt.pause(0.05)
    #     plt.cla()
    #     plt.draw()


import numpy as np
from random import sample
from masced_bandits.bandits.Expert import Expert
from masced_bandits.bandits.Bandit import Bandit
from masced_bandits.bandit_options import bandit_args
from masced_bandits.bandits.EXP3 import EXP3
from masced_bandits.bandits.EwS import EwS
ACTION = 0
REWARD = 1
N_K = 2

TOTAL_ROUNDS = 1#round(3000 / 60)

"""
Requires the following arguments:
learning_rate
num_experts
expert
"""
class EXP4(Bandit):
    def __init__(self, **kwargs):
        super().__init__("EXP4-" + str(kwargs))
        
        self.num_exps = kwargs["num_experts"]
        
        self.expert = self.expert_to_class(kwargs["expert"])

        self.distribution = None
        
        self.eta = float(kwargs["learning_rate"])

        self.experts = []

        self.knowledge = None
        self.previous_expert = 0 #this can be any expert but needs to be specified.
       
        self.distribution = [1.0/self.num_exps] * self.num_exps

        for i in range(self.num_exps):
            exp_instance = self.expert("FH")
            #exp_instance.eta = 0.1#np.sqrt( np.log(len(self.arms))/ (TOTAL_ROUNDS * len(self.arms))   )

            if(bandit_args["preload_knowledge"]):
                
                if(isinstance(exp_instance,EwS)):
                    exp_instance.weights = bandit_args["expert_preknowledge"][i][0]
                    exp_instance.arm_reward_pairs = bandit_args["expert_preknowledge"][i][1]
                else:
                    exp_instance.weights = bandit_args["expert_preknowledge"][i]
            exp_instance.distr_func()
            self.experts.append(exp_instance)


    
    def get_next_arm(self, reward):

        experts_matrix = np.matrix([expert.distribution for expert in self.experts])
        dist_over_arms = self.distribution * experts_matrix

        dist_over_arms = list(np.array(dist_over_arms).flatten())

        approx_arm_rewards = []
        for i in range(len(self.arms)):
            approx_rew = 1 - ( (1 if self.arms[i] == self.last_action else 0) /dist_over_arms[i]) * (1 - reward)
            approx_arm_rewards.append(approx_rew)
        
        experts_weights = list(np.array(np.matmul(experts_matrix,approx_arm_rewards)).flatten())
   
        #self.experts[self.previous_expert].propagate_reward(reward, self.arms.index(self.last_action)) this makes the expert learn


        sum_prev_weights = sum([ np.exp(self.eta * experts_weights[j])* self.distribution[j] for j in range(self.num_exps)])

        for odd_index, expert_odd in enumerate(self.distribution):
            adjusted_weight = np.exp(self.eta * experts_weights[odd_index])
            
            self.distribution[odd_index] = ((adjusted_weight * expert_odd) / sum_prev_weights)

 
        expert_choice = np.random.choice(np.arange(0, self.num_exps), p= self.distribution) #first choose the expert

        chosen_action = self.arms[self.experts[expert_choice].sample_action()]  #get action from that expert
        
        self.last_action = chosen_action
        self.previous_expert = expert_choice
       
        return chosen_action
    
    def expert_to_class(self, choice):
        funcs = {
                "EXP3": EXP3,
                "EwS" : EwS
            }
            
        func = funcs.get(choice)
        ##print(func.__doc__)
        return func

    def expert_status(self):
        for i, expert in enumerate(self.experts):
            print("-----EXPERT" + str(i) +"---")
            print('Distribution: ' + str(expert.distribution))
            print('Weights: ' + str(expert.weights))
            print("-----END EXPERT---")





import pickle
from masced_bandits.bandit_options import bandit_args


OPT_REVENUE = 1.5
BASIC_REVENUE = 1
PRECISION = 10**(-5)
SERVER_COST = 10
RT_THRESH = 0.75
MAX_SERVICE_RATE = 1 / 0.4452713


def assign_utilityfunc():
	func =  {
        "SEAMS2017A": utilitySEAMS2017A,
        "SEAMS2022": utilitySEAMS2022,
    }.get(bandit_args["utility_function"])

	return func
    
    
def say_hi(argument):
	print("Hello! I'm totally with new addition bandits and " + str(argument))
	return "this is the string I returned from say_hi"

def save_to_pickle(data_to_pkl, name):
	"""Takes any variable and a name for the pickle file and pickles it"""
	output = open(str(name) + '.pkl', 'ab+')
	pickle.dump(data_to_pkl, output)
	output.close()

def load_from_pickle(name):
	pkl_file = open(name + ".pkl", 'rb')
	data = pickle.load(pkl_file)
	pkl_file.close()
	return data

def pickle_test(data_to_pkl):
	print("pickling..")
	try:
		#read in data
		pkl_file = open('knowledge.pkl', 'rb')
		knowledge = pickle.load(pkl_file)
		print(knowledge)
		pkl_file.close()

		#adjust data
		knowledge = [item+1 for item in knowledge]
		output = open('knowledge.pkl', 'wb')
		pickle.dump(knowledge, output)
		output.close()
		return
	except IOError:
		#initial time
		initial_data = [0,0,0]
		output = open('knowledge.pkl', 'wb')
		pickle.dump(initial_data, output)
		output.close()
		return

def truncate(utility):
	bounds = bandit_args["bounds"]

	lower_bound, upper_bound = bounds

	old_range = upper_bound - lower_bound
	out_of_bounds = False


	if(utility > upper_bound):
		if(bandit_args["dynamic_bounds"]): 
			upper_bound = utility
			out_of_bounds = True
		else:
			utility = upper_bound
			
	elif(utility < lower_bound):
		if(bandit_args["dynamic_bounds"]):
			lower_bound = utility
			out_of_bounds = True
		else:
			utility = lower_bound

	
	
	new_range = upper_bound - lower_bound

	result = float((utility - lower_bound)/new_range)

	if(out_of_bounds):
		bandit_args["bounds"] = (lower_bound, upper_bound)
		return result, True, old_range/new_range
	else: return result, False, None

def convert_conf(new_conf, current_conf):
	#if current (2, 1.0) and new pair is (3,1.0) return "add_server"

	if(new_conf == current_conf): return "do nothing"

	server_difference = new_conf[0] - current_conf[0]

	commands = ['add_server'] * int(server_difference)

	if(not commands): commands = ["remove_server"] * int(abs(server_difference))

	if(new_conf[1] != current_conf[1]):
		commands.append("set_dimmer " + str(new_conf[1]))
	
	return commands

def utilitySEAMS2017A(arrival_rate, dimmer, avg_response_time, max_servers, servers, truncate=True):
	ur_opt = arrival_rate * OPT_REVENUE
	ur = arrival_rate * ((1-dimmer) * BASIC_REVENUE + (dimmer * OPT_REVENUE))
	#ur = arrival_rate * ( (1-dimmer) * BASIC_REVENUE + dimmer * OPT_REVENUE)
	
	bounds = bandit_args["bounds"]

	if((avg_response_time <= RT_THRESH) and (ur >= ur_opt - PRECISION)):
		uc = SERVER_COST * (max_servers - servers)
		return [truncate(ur + uc, bounds)]
	else:
		if(avg_response_time <= RT_THRESH):
			if(truncate):return [truncate(ur, bounds)]
			else: return[ur]
		else:
			max_throughput = max_servers * MAX_SERVICE_RATE
			if(truncate): return [truncate(min(0, arrival_rate - max_throughput) * OPT_REVENUE, bounds)]
			else: return [min(0, arrival_rate - max_throughput) * OPT_REVENUE]


def utilitySEAMS2022(arrival_rate, dimmer, avg_response_time, max_servers, servers, doTruncate=True):
	ur = arrival_rate * ((1 - dimmer) * BASIC_REVENUE + dimmer * OPT_REVENUE)
	uc = SERVER_COST * (max_servers - servers)
	urt = 1 - ((avg_response_time-RT_THRESH)/RT_THRESH)
	
	
	UPPER_RT_THRESHOLD = RT_THRESH * 4

	delta_threshold = UPPER_RT_THRESHOLD-RT_THRESH

	UrtPosFct = (delta_threshold/RT_THRESH) 

	urt = None
	if(avg_response_time <= UPPER_RT_THRESHOLD):
		urt = ((RT_THRESH - avg_response_time)/RT_THRESH) 
	else: 
		urt = ((RT_THRESH - UPPER_RT_THRESHOLD)/RT_THRESH)

	urt_final = None
	if(avg_response_time <= RT_THRESH):
		urt_final = urt*UrtPosFct 
	else:
		urt_final = urt

	revenue_weight = 0.7
	server_weight = 0.3
	utility = urt_final*((revenue_weight*ur)+(server_weight*uc))
	
	if(doTruncate): 
		truncated_reward, is_bound_diff, bound_delta = truncate(utility)
		return [truncated_reward], is_bound_diff, bound_delta
	else: return [utility], False, None
   

calculate_utility = assign_utilityfunc()



import numpy as np
from random import sample
from masced_bandits.bandit_options import bandit_args
from masced_bandits.utilities import convert_conf, save_to_pickle, calculate_utility
from masced_bandits.bandits.Bandit import Bandit
from masced_bandits.bandits.Expert import Expert
from statistics import mean

ACTION = 0
REWARD = 1
N_K = 2

#ETA = 1

#EXP3 a.k.a. SoftMax
class EXP3(Bandit, Expert): 
    def __init__(self, **kwargs): 
        super().__init__("EXP3-" + str(kwargs))
        self.weights, self.distribution = self.exp3_initialize(len(self.arms))
        self.num_arms = len(self.arms)
 
        if("horizon" in kwargs):
            self.eta = np.sqrt(np.log(len(self.arms)) / (len(self.arms) * int(kwargs["horizon"])) ) #0.1
        elif("learning_rate" in kwargs):
            self.eta = float(kwargs["learning_rate"])
        else:
            raise RuntimeError('EXP3 not hyperparamterized')
        
        self.distr_func()

        
       
    def exp3_initialize(self, num_arms):
        return [0] * num_arms, []

    
    def get_next_arm(self, reward):

        self.update_func(reward, self.arms.index(self.last_action)) #Update weights

        self.distr_func() #(re-)calculate Pt
        
        new_action = self.sample_action()
  
        self.last_action = self.arms[new_action]

        return self.arms[new_action]

    def propagate_reward(self, reward, chosen_action):
        self.update_func(reward, chosen_action)

        self.distr_func()


    def distr_func(self):
        # exp(eta * S^_t-1i) / SUMkj=1 exp(eta * S^_t-1j)

        sum_weights = sum([np.exp(self.eta * weight) for weight in self.weights])

        self.distribution.clear()
        #P_t = 
        self.distribution.extend([np.exp(self.eta * weight)/sum_weights for weight in self.weights])

    def update_func(self, payoff, action):
        #S^_ti = S^_t-1i + 1 - I{A_t = i}(1 - X_t) / P_ti
        for weight_i in range(len(self.weights)):
            if(weight_i == action):
                self.weights[action] = self.weights[action] + 1 - ((1-payoff)/self.distribution[action]) 
            else:
                self.weights[weight_i] = self.weights[weight_i] + 1
        return



import numpy as np
from random import sample
from masced_bandits.bandit_options import bandit_args
from masced_bandits.utilities import convert_conf, save_to_pickle, calculate_utility
from masced_bandits.bandits.Bandit import Bandit
from masced_bandits.bandits.Expert import Expert
from statistics import mean

ACTION = 0
REWARD = 1
N_K = 2

#ETA = 1


class EXP3S(Bandit, Expert):
    def __init__(self, **kwargs): 
        super().__init__("EXP3S-" + str(kwargs))
        self.weights, self.distribution = self.exp3s_initialize(len(self.arms))
        self.num_arms = len(self.arms)

        #np.random.seed(1337)
        total_count = kwargs["horizon"] 
        self.gamma = 1/total_count
        self.alpha = min(1, np.sqrt( (self.num_arms * np.log(self.num_arms * total_count)) / total_count))
        #alpha is the learning rate, the lower it is the harder the probabilities commit. If you set it too high you will get overflowing weights
        #gamma is the discount factor
        #low gamma entails your new reward has slight influence on weights but the weight affects the dsitribution heavily
        #high gamma entails your new reward influences the weights but the weights don't affect the distribution
        self.distr_func()

        
        
       
    def exp3s_initialize(self, num_arms):
        return [1] * num_arms, []



    def get_next_arm(self, reward):
        #print("received this " + str(reward))

        #print("my distribution is ")
        #print(self.distribution)    
        
        self.update_func(reward, self.last_action) #Update weights

        #print("now my weights are")
        #print(self.weights)
        self.distr_func() #(re-)calculate Pt
        
        #print("now my distribution is ")
        #print(self.distribution)     

        new_action = self.sample_action()
  
        self.last_action = self.arms[new_action]

        return self.arms[new_action]

    def propagate_reward(self, reward, chosen_action):
        self.update_func(reward, chosen_action)

        self.distr_func()

    def distr_func(self):
        # exp(eta * S^_t-1i) / SUMkj=1 exp(eta * S^_t-1j)

        sum_weights = sum(self.weights)


        self.distribution.clear()
        #P_t = 
        self.distribution.extend([(1 - self.gamma) * (weight/sum_weights) + (self.gamma/self.num_arms) for weight in self.weights])


    def update_func(self, reward, chosen_action):
        # print("making new weights")
        # print("the distribution is")
        # print(self.distribution)
        reward_estimates = [0] * len(self.arms)

        chosen_arm_i = self.arms.index(chosen_action)

        reward_estimates[chosen_arm_i] = reward/self.distribution[chosen_arm_i]
        # print("rew estimates")
        # print(reward_estimates)
        sum_prev_weights = sum(self.weights)
        #print("prev weight sum")
        #print(sum_prev_weights)

        #print("prev weights are")
        #print(self.weights)
        for weight_i in range(len(self.weights)):
            prev_weight = self.weights[weight_i]
            #print("gamma is " + str(self.gamma))
            leftside = prev_weight * np.exp(self.gamma * reward_estimates[weight_i] / self.num_arms)
            #print("left side is ")
            #print(leftside)
            rightside = ((np.exp(1) * self.alpha) / self.num_arms) * sum_prev_weights
            #print("alpha is ")
            #print(self.alpha)
            #print("right side is ")
            #print(rightside)
            self.weights[weight_i] = leftside + rightside #new weight






import numpy as np
import time
from masced_bandits.utilities import calculate_utility, convert_conf
from masced_bandits.bandit_options import bandit_args
from masced_bandits.bandits.Bandit import Bandit



CUM_REWARD = 0
N_K = 1
class ETC(Bandit):
    def __init__(self, **kwargs):
        super().__init__("ETC")

        self.bandit_round = 0

        self.explore_factor = float(kwargs.get('exploration_rounds',10))

        self.arm_reward_pairs = {}
        for arm in self.arms: self.arm_reward_pairs[arm] = [0.0,0.0]

    def get_next_arm(self, reward):
        self.bandit_round+=1
        self.arm_reward_pairs[self.last_action][CUM_REWARD]+=reward
        self.arm_reward_pairs[self.last_action][N_K]+=1

        if  self.bandit_round <= ((len(self.arms)-1) * self.explore_factor):
            new_action = self.arms[(self.bandit_round % len(self.arms))]
        else: 
            new_action = max(self.arms, key=lambda arm: self.reward_average(arm))

        self.last_action = new_action
        return new_action


    def reward_average(self, arm):
        return self.arm_reward_pairs[arm][CUM_REWARD] / self.arm_reward_pairs[arm][N_K] if self.arm_reward_pairs[arm][N_K] else 0


    

    





import random
import time
import itertools
bandit_args = {
    "start_time": round(time.time()),
    "arms": [],#[(4,1.0), (5,1.0), (10,1.0)],
    'cleaning': False,
    'stored_choice': None,
    "bandit_instance": None,
    "initial_configuration": None, #(4, 1.0),
    "shuffle": False,
    "record_decisions": False,
    "bounds": (0,1), #-400,300
    "utility_function": "SEAMS2022",
    "number_of_experts": 2,
    "preload_knowledge": False,
    "expert": "EXP3",
    "round_counter": [0,0],
    "dynamic_bounds": False,
    "expert_preknowledge": [[36.608650878493556, 101.58972407707387, 155.44454636285019, 158.48317206373139, 160.14239383167592], 
                            [6.5191987585019859, 19.81411169769218, 41.438958866194355, -18.95283084856483, 126.99528063289736]]
}

if(bandit_args["shuffle"]) :random.shuffle(bandit_args["arms"])


def initialize_arguments(arms, initial_arm_index, shuffle=False, record_decisions=False, bounds=(0,1), dynamic_bounds=False):
    args = locals()

    for key in args.keys():
        bandit_args[key] = args[key]


    bandit_args["initial_configuration"] = args["arms"][initial_arm_index]



import numpy as np
from masced_bandits.bandits.Bandit import Bandit

CUM_REWARD = 0
CUM_SQ_REWARD = 1
N_K = 2

class UCBImproved(Bandit):
    def __init__(self, **kwargs):
        super().__init__("UCB-Improved")
        self.horizon = int(kwargs.get("horizon",100))
        self.removable_arms = [arm for arm in self.arms]
        self.arm_reward_pairs = {}
        for arm in self.arms: self.arm_reward_pairs[arm] = [0.0,0.0,0.0]
        self.delta_m = 1.0

        
        
    def get_next_arm(self, reward):
   
        if(len(self.removable_arms) == 1): 
            #converged
            return self.removable_arms[0]
        
        self.arm_reward_pairs[self.last_action][CUM_REWARD]+=reward
        self.arm_reward_pairs[self.last_action][CUM_SQ_REWARD]+=np.square(reward)
        self.arm_reward_pairs[self.last_action][N_K]+=1
        
        delta_sq = np.square(self.delta_m)

        n_m = np.ceil( (2 * np.log(self.horizon * delta_sq))/ delta_sq )

        for arm in self.removable_arms:
            if self.arm_reward_pairs[arm][N_K] < n_m: 
                #exploring
                self.last_action = arm
                return arm

        fac = np.sqrt(np.log(self.horizon * delta_sq) / (2 * n_m))
    
        pair_avgs = [self.arm_reward_pairs[arm][CUM_REWARD]/self.arm_reward_pairs[arm][N_K] \
                    for arm in self.removable_arms]

        del_boundary = max([pair_avg - fac for pair_avg in pair_avgs])

        del_candidates = []

        for avg_i, arm_avg in enumerate(pair_avgs):

            
            if (arm_avg + fac) < del_boundary:
                del_candidates.append(self.removable_arms[avg_i])

        for candidate in del_candidates: self.removable_arms.remove(candidate) 

        self.delta_m = self.delta_m/2

        return self.get_next_arm(reward)